@@ -25,6 +25,10 @@
 
 #include "gclue-location.h"
 #include <math.h>
+#include <string.h>
+#include <stdlib.h>
+
+#define TIME_DIFF_THRESHOLD 60
 
 struct _GClueLocationPrivate {
         gdouble speed;
@@ -156,6 +160,142 @@ gclue_location_init (GClueLocation *location)
         location->priv->heading = GCLUE_LOCATION_HEADING_UNKNOWN;
 }
 
+static gdouble
+get_accuracy_from_hdop (gdouble hdop)
+{
+        /* FIXME: These are really just rough estimates based on:
+         *        http://en.wikipedia.org/wiki/Dilution_of_precision_%28GPS%29#Meaning_of_DOP_Values
+         */
+        if (hdop <= 1)
+                return 0;
+        else if (hdop <= 2)
+                return 1;
+        else if (hdop <= 5)
+                return 3;
+        else if (hdop <= 10)
+                return 50;
+        else if (hdop <= 20)
+                return 100;
+        else
+                return 300;
+}
+
+static gdouble
+parse_coordinate_string (const char *coordinate,
+                         const char *direction)
+{
+        gdouble minutes, degrees, out;
+        gchar *degrees_str;
+        gchar *dot_str;
+        gint dot_offset;
+
+        if (coordinate[0] == '\0' ||
+            direction[0] == '\0' ||
+            direction[0] == '\0')
+                return INVALID_COORDINATE;
+
+        if (direction[0] != 'N' &&
+            direction[0] != 'S' &&
+            direction[0] != 'E' &&
+            direction[0] != 'W') {
+                g_warning ("Unknown direction '%s' for coordinates, ignoring..",
+                           direction);
+                return INVALID_COORDINATE;
+        }
+
+        dot_str = g_strstr_len (coordinate, 6, ".");
+        if (dot_str == NULL)
+                return INVALID_COORDINATE;
+        dot_offset = dot_str - coordinate;
+
+        degrees_str = g_strndup (coordinate, dot_offset - 2);
+        degrees = g_ascii_strtod (degrees_str, NULL);
+        g_free (degrees_str);
+
+        minutes = g_ascii_strtod (dot_str - 2, NULL);
+
+        /* Include the minutes as part of the degrees */
+        out = degrees + (minutes / 60.0);
+
+        if (direction[0] == 'S' || direction[0] == 'W')
+                out = 0 - out;
+
+        return out;
+}
+
+static gdouble
+parse_altitude_string (const char *altitude,
+                       const char *unit)
+{
+        if (altitude[0] == '\0' || unit[0] == '\0')
+                return GEOCODE_LOCATION_ALTITUDE_UNKNOWN;
+
+        if (unit[0] != 'M') {
+                g_warning ("Unknown unit '%s' for altitude, ignoring..",
+                           unit);
+
+                return GEOCODE_LOCATION_ALTITUDE_UNKNOWN;
+        }
+
+        return g_ascii_strtod (altitude, NULL);
+}
+
+static gint64
+parse_nmea_timestamp (const char *nmea_ts)
+{
+        char parts[3][3];
+        int i, hours, minutes, seconds;
+        GDateTime *now, *ts = NULL;
+        guint64 ret;
+
+        now = g_date_time_new_now_utc ();
+        ret = g_date_time_to_unix (now);
+
+        if (strlen (nmea_ts) < 6) {
+                if (strlen (nmea_ts) >= 1)
+                        /* Empty string just means no ts, so no warning */
+                        g_warning ("Failed to parse NMEA timestamp '%s'",
+                                   nmea_ts);
+
+                goto parse_error;
+        }
+
+        for (i = 0; i < 3; i++) {
+                memmove (parts[i], nmea_ts + (i * 2), 2);
+                parts[i][2] = '\0';
+        }
+        hours = atoi (parts[0]);
+        minutes = atoi (parts[1]);
+        seconds = atoi (parts[2]);
+
+        ts = g_date_time_new_utc (g_date_time_get_year (now),
+                                  g_date_time_get_month (now),
+                                  g_date_time_get_day_of_month (now),
+                                  hours,
+                                  minutes,
+                                  seconds);
+
+        if (g_date_time_difference (ts, now) > TIME_DIFF_THRESHOLD) {
+                g_debug ("NMEA timestamp '%s' in future. Assuming yesterday's.",
+                         nmea_ts);
+                g_date_time_unref (ts);
+
+                ts = g_date_time_new_utc (g_date_time_get_year (now),
+                                          g_date_time_get_month (now),
+                                          g_date_time_get_day_of_month (now) - 1,
+                                          hours,
+                                          minutes,
+                                          seconds);
+        }
+
+        ret = g_date_time_to_unix (ts);
+        g_date_time_unref (ts);
+parse_error:
+        g_date_time_unref (now);
+
+        return ret;
+}
+
 /**
  * gclue_location_new:
  * @latitude: a valid latitude
@@ -179,10 +319,14 @@ gclue_location_new (gdouble latitude,
 }
 
 /**
- * gclue_location_new_with_description:
+ * gclue_location_new_full:
  * @latitude: a valid latitude
  * @longitude: a valid longitude
  * @accuracy: accuracy of location in meters
+ * @speed: speed in meters per second
+ * @heading: heading in degrees
+ * @altitude: altitude of location in meters
+ * @timestamp: timestamp in seconds since the Epoch
  * @description: a description for the location
  *
  * Creates a new #GClueLocation object.
@@ -190,19 +334,117 @@ gclue_location_new (gdouble latitude,
  * Returns: a new #GClueLocation object. Use g_object_unref() when done.
  **/
 GClueLocation *
-gclue_location_new_with_description (gdouble     latitude,
-                                     gdouble     longitude,
-                                     gdouble     accuracy,
-                                     const char *description)
+gclue_location_new_full (gdouble     latitude,
+                         gdouble     longitude,
+                         gdouble     accuracy,
+                         gdouble     speed,
+                         gdouble     heading,
+                         gdouble     altitude,
+                         guint64     timestamp,
+                         const char *description)
 {
         return g_object_new (GCLUE_TYPE_LOCATION,
                              "latitude", latitude,
                              "longitude", longitude,
                              "accuracy", accuracy,
+                             "speed", speed,
+                             "heading", heading,
+                             "altitude", altitude,
+                             "timestamp", timestamp,
                              "description", description,
                              NULL);
 }
 
+/**
+ * gclue_location_create_from_gga:
+ * @gga: NMEA GGA sentence
+ * @error: Place-holder for errors.
+ *
+ * Creates a new #GClueLocation object from a GGA sentence.
+ *
+ * Returns: a new #GClueLocation object, or %NULL on error. Unref using
+ * #g_object_unref() when done with it.
+ **/
+GClueLocation *
+gclue_location_create_from_gga (const char *gga, GError **error)
+{
+        GClueLocation *location = NULL;
+        gdouble latitude, longitude, accuracy, altitude;
+        gdouble hdop; /* Horizontal Dilution Of Precision */
+        guint64 timestamp;
+        char **parts;
+
+        parts = g_strsplit (gga, ",", -1);
+        if (g_strv_length (parts) < 14) {
+                g_set_error_literal (error,
+                                     G_IO_ERROR,
+                                     G_IO_ERROR_INVALID_ARGUMENT,
+                                     "Invalid NMEA GGA sentence");
+                goto out;
+        }
+
+        /* For syntax of GGA sentences:
+         * http://www.gpsinformation.org/dale/nmea.htm#GGA
+         */
+        timestamp = parse_nmea_timestamp (parts[1]);
+        latitude = parse_coordinate_string (parts[2], parts[3]);
+        longitude = parse_coordinate_string (parts[4], parts[5]);
+        if (latitude == INVALID_COORDINATE || longitude == INVALID_COORDINATE) {
+                g_set_error_literal (error,
+                                     G_IO_ERROR,
+                                     G_IO_ERROR_INVALID_ARGUMENT,
+                                     "Invalid NMEA GGA sentence");
+                goto out;
+        }
+
+        altitude = parse_altitude_string (parts[9], parts[10]);
+        if (altitude == GEOCODE_LOCATION_ALTITUDE_UNKNOWN)
+                goto out;
+
+        hdop = g_ascii_strtod (parts[8], NULL);
+        accuracy = get_accuracy_from_hdop (hdop);
+
+        location = g_object_new (GCLUE_TYPE_LOCATION,
+                                 "latitude", latitude,
+                                 "longitude", longitude,
+                                 "accuracy", accuracy,
+                                 "timestamp", timestamp,
+                                 NULL);
+        if (altitude != GEOCODE_LOCATION_ALTITUDE_UNKNOWN)
+                g_object_set (location, "altitude", altitude, NULL);
+
+out:
+        g_strfreev (parts);
+        return location;
+}
+
+/**
+ * gclue_location_duplicate:
+ * @location: the #GClueLocation instance to duplicate.
+ *
+ * Creates a new copy of @location object.
+ *
+ * Returns: a new #GClueLocation object. Use g_object_unref() when done.
+ **/
+GClueLocation *
+gclue_location_duplicate (GClueLocation *location)
+{
+        GeocodeLocation *gloc = GEOCODE_LOCATION (location);
+
+        return g_object_new
+                (GCLUE_TYPE_LOCATION,
+                 "description", geocode_location_get_description (gloc),
+                 "latitude", geocode_location_get_latitude (gloc),
+                 "longitude", geocode_location_get_longitude (gloc),
+                 "accuracy", geocode_location_get_accuracy (gloc),
+                 "altitude", geocode_location_get_altitude (gloc),
+                 "crs", geocode_location_get_crs (gloc),
+                 "timestamp", geocode_location_get_timestamp (gloc),
+                 "speed", location->priv->speed,
+                 "heading", location->priv->heading,
+                 NULL);
+}
+
 /**
  * gclue_location_get_speed:
  * @location: a #GClueLocation
@@ -257,9 +499,9 @@ gclue_location_set_speed_from_prev_location (GClueLocation *location,
                           GCLUE_IS_LOCATION (prev_location));
 
         if (prev_location == NULL) {
-               location->priv->speed = GCLUE_LOCATION_SPEED_UNKNOWN;
+               speed = GCLUE_LOCATION_SPEED_UNKNOWN;
 
-               return;
+               goto out;
         }
 
         gloc = GEOCODE_LOCATION (location);
@@ -268,11 +510,16 @@ gclue_location_set_speed_from_prev_location (GClueLocation *location,
         timestamp = geocode_location_get_timestamp (gloc);
         prev_timestamp = geocode_location_get_timestamp (prev_gloc);
 
-        g_return_if_fail (timestamp != prev_timestamp);
+        if (timestamp <= prev_timestamp) {
+               speed = GCLUE_LOCATION_SPEED_UNKNOWN;
+
+               goto out;
+        }
 
         speed = geocode_location_get_distance_from (gloc, prev_gloc) *
                 1000.0 / (timestamp - prev_timestamp);
 
+out:
         location->priv->speed = speed;
 
         g_object_notify (G_OBJECT (location), "speed");
